/* main.c
   
   See the README file for usage instructions, and INSTALL
   for help on installing
   
*/


/*
  SquidClamav - A Clamav Antivirus Redirector for Squid

  This project is a modified version of the excellent Squirm Redirector
  for Squid Maintained by Chris Foote, and copyrighted
  as follow :

  Copyright (C) 1998 Chris Foote & Wayne Piekarski

  The original Squirm version used was squirm-1.0betaB. Some other parts
  are cut and paste from the ex1.c program given in the ClamAv distribution
  and are copyrighted: Copyright (C) 2002 - 2004 Tomasz Kojm

  All other code: Copyright (C) 2005-2009 Gilles Darold

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
    
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  Please see the file GPL in this directory for full copyright
  information.
*/

#define TRUE  1
#define FALSE 0
#define FSIZE 102400		/* scan max 100kb */
#define MAXREDIR 10

#include "log.h"
#include "util.h"
#include "squidclamav.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sysexits.h>
#include <errno.h>
#include <curl/curl.h>
#include <signal.h>
#include <sys/time.h>

static char * escape_quote (char *s);
void timeit (struct timeval start, char *level);
int dconnect (void);
void replace_chr(char string[], char *from, char *to);
char *replace_str(char *str, char *orig, char *rep);
void free_global ();		/* routine to free global pointer */
void freeBuff (struct IN_BUFF *);
/* routine to cleanup on signal */
void sighup ();
void sigterm ();
void sigusr1();
int purge_cache(char *url);
int connectINET(char *serverHost, int serverPort);

int bridge_mode = 0;
int interactive = 0;
int check_content_type = 0;
int has_regex = 0;
int debug = 0;
int force = 0;
int statit = 0;
int timeout = 10;
int wsockd = -1;
char *alternate_config = NULL;
char *redirect_url = NULL;
char *squidguard = NULL;
char *log_file = NULL;
char *proxy = NULL;
char *clamd_local = NULL;
char *clamd_ip = NULL;
char *clamd_port = NULL;
char *useragent = NULL;
FILE *sgfpw = NULL;
FILE *sgfpr = NULL;
int maxredir = MAXREDIR;
int trust_cache = 0;
char *squid_ip = NULL;
char *squid_port = NULL;

regex_t *regexv;	/* Array of compiled patterns */
CURL *eh = NULL;	/* libcurl handler */
struct curl_slist *slist = NULL; 

int skip_scan = 0;

void
sigalarm (int sig)
{
  logit(log_file, "LOG Timeout waiting for clamd response\n");
  exit (3);
}

/* Get download as stream  */
size_t
write_data (void *ptr, size_t data_size, size_t nmemb, void *stream)
{
  int ret;
  int msize = data_size * nmemb;

  if (debug > 1)
    logit(log_file, "DEBUG Sending data to clamd\n");
  ret = write (wsockd, ptr, msize);
  if (ret <= 0)
    {
	    logit(log_file, "ERROR Can't write to clamd socket.\n");
	    close (wsockd);
    }
  else
    {
      if (debug > 1)
	logit(log_file, "DEBUG Write %d bytes on %d to socket\n", ret, msize);
    }

  return msize;
}

/* Get squid header as stream  */
size_t
read_data (void *ptr, size_t data_size, size_t nmemb, void *stream)
{
  int ret;
  int msize = data_size * nmemb;

  if (debug > 2)
    logit(log_file, "DEBUG Received HTTP-HEADER: %s", ptr);
  /* Detect if the file come from Squid cache and if we may trust it*/
  if ((trust_cache == 1) && (strncmp(ptr, "X-Cache: HIT", 12) == 0) ) {
    skip_scan = 1;
  }

  return msize;
}

/* 
    Under normal conditions, SquidClamav will never abort (you don't
    want your Squid to suddenly die if you've put an invalid
    regex into your config file, or renamed the config file!  
    Under these abnormal conditions, squidclamav will run in bridge_mode 
    where stdin is echoed to stdout!

    If you've entered an invalid regex expression that regcomp()
    doesn't like, or perhaps an invalid network number, check it 
    out in the ERROR_LOG.
    
    It should never ever abort under normal circumstances - even
    if it can't allocate memory (which shouldn't happen, and Squid 
    would be running like a dog under low memory conditions anyway) 
    bridge_mode will be set and it will keep running :-)
    
*/

int
main (int argc, char **argv)
{
  char *sbuff = NULL; /* Squid entry */
  int buff_status = 0;
  char *command = NULL; /* Chained system command */
  char cbuff[MAX_BUFF]; /* Clamd return */
  char *escaped = NULL;

  double usize = 0;
  char *content_type = NULL;
  char err[CURL_ERROR_SIZE];
  int found = 0;
  int virusfound = 0;

  /* Clamd socket variable */
  char urlredir[MAX_BUFF];
  int nbread = 0;
  int sockd;
  int loopw = 60, port;
  struct sockaddr_in server;
  struct sockaddr_in peer;
  int peer_size;
  char *pt = NULL;
  int force_scan = 0;

  struct timeval g_start; /* Start global time structure */
  struct timeval cs_start; /* Start chaining and scanning time structure */
  struct timespec sleeptime;
  sleeptime.tv_sec = 0;
  sleeptime.tv_nsec = 1000000;
 
  FILE *chain_pipe = NULL;
  char *chain_ret = NULL;

  int pipe1[2];
  int pipe2[2];
  pid_t pid;
  int usepipe = 0;

  /* go into interactive mode if we're run as root */
  if ((int) getuid () == 0)
    {
      interactive = 1;
      fprintf (stderr,
	       "SquidClamav running as UID 0: writing logs to stderr\n");
    }

  /* check for alternate config file given as first argument */
  if (argc == 2) {
	alternate_config = (char *)malloc(sizeof(char) * LOW_BUFF);
	if(alternate_config == NULL) {
		fprintf(stderr, "unable to allocate memory in main()\n");
		exit(EXIT_FAILURE);
	}
	xstrncpy(alternate_config, argv[1], LOW_BUFF);

  }

  if (argc > 2)
    {
      fprintf (stderr, "squidclamav: invalid arguments\n");
      fprintf (stderr, "Usage: %s [squidclamav.conf]\n", argv[0]);
      nanosleep(&sleeptime, NULL);		/* don't do DoS */
      exit(0);
    }

  /* make stdout line buffered */
  if (setvbuf (stdout, NULL, _IOLBF, 0) != 0)
    {
      fprintf (stderr, "stdout not line buffered exiting\n");
      nanosleep(&sleeptime, NULL);		/* don't do DoS */
      exit (2);
    }

  bridge_mode = 0;

 /*********************
   read config files
  ********************/
  load_patterns(0);
  regexv = regcomp_pattern();

  signal (SIGHUP, sighup);	/* set function calls on signal trap */
  signal (SIGTERM, sigterm);
  signal (SIGUSR1, sigusr1);

  if (bridge_mode)
    logit(log_file, "PANIC Invalid condition - continuing in BRIDGE mode\n");

  logit(log_file, "LOG SquidClamav (PID %d) started\n", (int) getpid ());

  if (!bridge_mode) {

      /* curl init */
      curl_global_init (CURL_GLOBAL_ALL);
      /* get an easy handle */
      if ((eh = curl_easy_init ()) == NULL)
	{
	  logit(log_file, "PANIC Curl easy initialization failed\n");
	  curl_global_cleanup ();
	  bridge_mode = 1;
	  logit(log_file, "PANIC Invalid condition - continuing in BRIDGE mode\n");
	}
      else
	{
	  /* set the curl output header redirected to a function */
	  curl_easy_setopt (eh, CURLOPT_HEADERFUNCTION, &read_data);
	  /* set the curl output to temporary file */
	  curl_easy_setopt (eh, CURLOPT_WRITEFUNCTION, &write_data);
	  /* set the error buffer */
	  curl_easy_setopt (eh, CURLOPT_ERRORBUFFER, err);
	  /* set the libcurl transfer timeout */
	  curl_easy_setopt (eh, CURLOPT_TIMEOUT, timeout);
	  /* do not install signal  handlers in thread context */
	  curl_easy_setopt (eh, CURLOPT_NOSIGNAL, 1);
	  /* Force HTTP 1.0 version */
	  curl_easy_setopt (eh, CURL_HTTP_VERSION_1_0, TRUE);
	  /* Do not fail on errors: use to prevent breaked download when
	   * size of the download is upper than clamd scan limit
	   * Drop error on debug mode  */
	  if (debug != 0) {
		  curl_easy_setopt (eh, CURLOPT_FAILONERROR, TRUE);
	  } else {
		  curl_easy_setopt (eh, CURLOPT_FAILONERROR, FALSE);
	  }
	  /* follow location (303 MOVED) */
	  curl_easy_setopt (eh, CURLOPT_FOLLOWLOCATION, TRUE);
	  /* only follow 3 redirects */
	  curl_easy_setopt (eh, CURLOPT_MAXREDIRS, maxredir);
	  /* simulates being another user agent */
	  if (useragent != NULL)
		curl_easy_setopt (eh, CURLOPT_USERAGENT, useragent);
	  /* overwrite the Pragma: no-cache HTTP header */
	  slist = curl_slist_append(slist, "pragma:");
	  curl_easy_setopt (eh, CURLOPT_HTTPHEADER, slist);
	}
   }
  /* create a squidguard child process and setup pipes */
  if ((bridge_mode == 0) && (squidguard != NULL)) {

	if ( pipe(pipe1) < 0  ||  pipe(pipe2) < 0 )
	{
		logit(log_file,"ERROR unable to open pipe, disabling squidGuard bidirectional pipe.\n");
	} else {
		if ( (pid = fork()) == -1) {
			logit(log_file,"ERROR unable to fork, disabling squidGuard bidirectional pipe.\n");
		} else {
			if(!pid) {
				close(pipe1[1]);
				dup2(pipe1[0],0);
				close(pipe2[0]);
				dup2(pipe2[1],1);
				execlp(squidguard,"squidGuard",0); /* run squidguard */
			} else {
				close(pipe1[0]);
				sgfpw = fdopen(pipe1[1], "w");
				if(!sgfpw) {
					logit(log_file,"ERROR unable to fopen squidGuard childs stdin, disabling squidGuard.\n");
				} else {
					sgfpr = fdopen(pipe2[0], "r");
					if(!sgfpr) {
						logit(log_file,"ERROR unable to fopen squidGuard childs stdout, disabling squidGuard.\n");
					}
				}
				logit(log_file,"bidirectional pipe to squidGuard childs ready...\n");
				usepipe = 1;
			}
		}
	}
  }

  /* Loop on squid inout */
  sbuff = (char *)malloc(sizeof(char)*MAX_BUFF); /* Squid entry */
  while (fgets(sbuff, MAX_BUFF, stdin) != NULL)
    {
	/* Initialize squid input buffer */
	struct IN_BUFF * in_buff;
	in_buff = ( struct IN_BUFF * )malloc( sizeof( struct IN_BUFF ) );
	if (in_buff == NULL) {
	    logit(log_file, "unable to allocate memory in main() in_buff\n");
	    /* do not go further */
	    continue;
	}

	/* chop sbuff */
	chomp(sbuff);

	/* Start timing */
	if (statit)
		gettimeofday(&g_start,NULL);

	/* separate the four fields
	 from the single input line 
	 of stdin */
	in_buff->url = (char * )malloc( sizeof(char)*MAX_BUFF);
	in_buff->src_address = (char * )malloc( sizeof(char)*MAX_CHAR);
	in_buff->ident = (char * )malloc( sizeof(char)*MAX_CHAR);
	in_buff->method = (char * )malloc( sizeof(char)*LOW_CHAR);
	buff_status = load_in_buff(sbuff, in_buff);

	/* chaining with SquidGuard - before bridge mode or not*/
	if ((bridge_mode == 0) && (squidguard != NULL)) {
		if (debug != 0)
		    logit(log_file, "DEBUG Sending request to chained program: %s\n", squidguard);
		if (usepipe == 0) {
			xfree(command);
			command = (char *)malloc(sizeof(char) * (MAX_BUFF * 2));
			if (command == NULL) {
				logit(log_file, "ERROR unable to allocate memory in main()\n");
				freeBuff(in_buff);
				/* do not go further */
				continue;
			}
			xstrncpy(command, "echo \"", LOW_CHAR);
			escaped = escape_quote(sbuff);
			strlcat(command, escaped, MAX_BUFF * 2);
			xfree(escaped);
			strlcat(command, "\" | ", MAX_BUFF * 2);
			strlcat(command, squidguard, MAX_BUFF * 2);
			if (debug > 1)
			    logit(log_file, "DEBUG Chained program system command execution: %s\n", command);
			chain_pipe = popen (command, "r");
			if ( !chain_pipe ) {
			    logit(log_file, "ERROR Can't open pipe to %s\n", squidguard);
			    /* go ahead */
			} else {
				/* the chained redirector must return empty line if ok
				   or the redirection url */
				chain_ret = (char *)malloc(sizeof(char)*MAX_BUFF);
				fgets(chain_ret, MAX_BUFF, chain_pipe);
				if (xstrnlen(chain_ret, MAX_BUFF) > 1) {
					if (debug > 1)
						logit(log_file, "DEBUG Received from chained program: %s", chain_ret);
					/* Close chain_pipe, checking for errors */
					if (pclose (chain_pipe) != 0)
						logit(log_file, "ERROR Could not run %s or other error.\n", squidguard);
					if (statit)
						timeit(g_start, "Chained program call");
					/* return the redirection url to squid */
					fprintf (stdout, "%s", chain_ret);
					/* do not go further */
					xfree(chain_ret);
					freeBuff(in_buff);
					continue;
				}
				xfree(chain_ret);
				/* Close chain_pipe, checking for errors */
				if (pclose (chain_pipe) != 0)
					logit(log_file, "ERROR Could not run %s or other error.\n", squidguard);
			}
		} else {
			escaped = escape_quote(sbuff);
			fprintf(sgfpw,"%s\n",escaped);
			xfree(escaped);
			fflush(sgfpw);
			/* the chained redirector must return empty line if ok or the redirection url */
			chain_ret = (char *)malloc(sizeof(char)*MAX_BUFF);
			fgets(chain_ret,MAX_BUFF,sgfpr);
			if (xstrnlen(chain_ret, MAX_BUFF) > 1) {
				if (debug > 1)
					logit(log_file, "DEBUG Received from chained program: %s", chain_ret);
				if (statit)
					timeit(g_start, "Chained program call");
				/* return the redirection url to squid */
				fprintf (stdout, "%s", chain_ret);
				/* do not go further */
				xfree(chain_ret);
				freeBuff(in_buff);
				continue;
			}
			xfree(chain_ret);
		}
		if (statit)
			timeit(g_start, "Chained program call");
	}

      /* if configs are completely invalid or some other
         exception occurs where we want the redirector to
         continue operation (so that Squid still works!),
         we simply echo stdin to stdout - i.e. "bridge mode" :-) */
      if (bridge_mode)
	{
	  fprintf (stdout, "\n");
          freeBuff(in_buff);
	  continue;
	}
     
      /* if four fields couldn't be separated, or the
         converted values aren't appropriate, then
         just echo back the line from stdin */

      if (buff_status == 1)
	{
	  fprintf (stdout, "\n");
          freeBuff(in_buff);
	  continue;
	}

      /* now we can compare the URL to virus scan rules */
      found = url_pattern_compare (regexv, in_buff->url);
      /* -1 mean ABORT or WHITELIST match => No antivir check */
      /* if there's a REGEX, no CONTENT pattern search and the regex
         pattern is not found => No antivir check */
      if ((found == -1)
	  || ((check_content_type == 0) && (has_regex == 1) && (found == 0)))
	{

	  if ( (debug > 1) && (found == -1) )
	    logit(log_file, "DEBUG No antivir check (ABORT or WHITELIST match) for url: %s\n", in_buff->url);
	  if ( (debug > 1) && (found != -1) )
	    logit(log_file, "DEBUG No antivir check (REGEX not found and no CONTENT search) for url: %s\n", in_buff->url);
	    if (statit)
		timeit(g_start, "Total");

	  /* no replacement for the URL was found */
	  fprintf (stdout, "\n");
          freeBuff(in_buff);
	  continue;

	} else {

	  skip_scan = 0;
	  /* set proxy */
	  if (strcmp(proxy,"none") != 0) {
		  if (debug > 0)
		    logit(log_file, "DEBUG Curl will use proxy: %s\n", proxy);
		curl_easy_setopt (eh, CURLOPT_PROXY, proxy);
	  }
	  /* set the url */
	  curl_easy_setopt (eh, CURLOPT_URL, in_buff->url);
	  /* only get the header to check size and content type */
	  curl_easy_setopt (eh, CURLOPT_NOBODY, TRUE);

	  /* download the header */
         if (debug > 1)
		logit(log_file, "DEBUG Getting Content-Type header for url %s\n", in_buff->url);
	  if (curl_easy_perform (eh) != 0)
	    {
	        if (force == 0) {
		      /* no header so just go back to squid */
		      if (debug != 0)
		        logit(log_file, "DEBUG bad header from url %s\n", in_buff->url);
			if (statit)
				timeit(g_start, "Total");
		      fprintf (stdout, "\n");
                      freeBuff(in_buff);
		      continue;
		} else {
		      if (debug != 0)
	                logit(log_file, "DEBUG Force scanning bad header from url %s\n", in_buff->url);
		    force_scan = 1;
		}
	    }
	  if (skip_scan == 1) {
		if (debug > 0)
		   logit(log_file, "DEBUG HIT Cache found, trusted cached = already scanned, skipping...\n");
	      if (statit)
		timeit(g_start, "Total");
	      fprintf (stdout, "\n");
	      freeBuff(in_buff);
	      continue;
	  }

	  if (force_scan == 0) {
	  /* get the Content Length from the header */
	  if (curl_easy_getinfo (eh, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &usize)
	      != CURLE_OK)
	    {
		      /* not ok --> squid should handle this */
		      logit(log_file, "ERROR No content length from url %s\n", in_buff->url);
		      if (statit)
				timeit(g_start, "Total");
		      fprintf (stdout, "\n");
                      freeBuff(in_buff);
		      continue;
	    }
	  if (debug > 1)
	    logit(log_file, "DEBUG File size is %'.2f\n", usize);

	  /* get the CONTENT_TYPE out of the header */
	  if (check_content_type == 1)
	    {
	      if (curl_easy_getinfo (eh, CURLINFO_CONTENT_TYPE, &content_type) != CURLE_OK)
		{
		  /* not ok --> squid should handle this */
		  logit(log_file, "ERROR No content type from url %s\n",
		       in_buff->url);
		  if (statit)
			timeit(g_start, "Total");
		  fprintf (stdout, "\n");
                  freeBuff(in_buff);
		  continue;
	      }
	      if ((content_type == NULL) || (content_pattern_compare(regexv, content_type) == 0))
		{
		  if (debug != 0)
		    logit(log_file, "DEBUG No antivir check (CONTENT not match) for content-type: %s\n", content_type);
			if (content_type != NULL)
				memset(content_type, 0, sizeof(content_type));
		    if (statit)
			timeit(g_start, "Total");
		  fprintf (stdout, "\n");
                  freeBuff(in_buff);
		  continue;
		}
	        if (debug != 0)
			logit(log_file, "DEBUG Content-Type: %s will be scanned\n", content_type);
		if (content_type != NULL)
			memset(content_type, 0, sizeof(content_type));
	    }
	}

	  /* Start timing */
	  if (statit)
		  gettimeofday(&cs_start,NULL);
	  /***** Open a socket to Clamd daemon *****/
	  if ((sockd = dconnect ()) < 0)
	    {
	      logit(log_file, "ERROR Can't connect to Clamd daemon\n");
	      fprintf (stdout, "\n");
              freeBuff(in_buff);
	      continue;
	    }
	  if (debug != 0)
	    logit(log_file, "DEBUG Sending STREAM to clamd.\n");

	  if (write (sockd, "STREAM", 6) <= 0)
	    {
	      logit(log_file, "ERROR Can't write to Clamd socket.\n");
	      fprintf (stdout, "\n");
              freeBuff(in_buff);
	      continue;
	    }

	  while (loopw > 0) {
	      char lbuff[1024];
	      memset(lbuff, 0, sizeof(lbuff));
	      read (sockd, lbuff, LOW_BUFF);
	      if ((pt = strstr (lbuff, "PORT")))
		{
		  pt += 5;
		  sscanf (pt, "%d", &port);
		  break;
		}
	      loopw--;
	    }
	  if (!loopw)
	    {
	      logit(log_file, "ERROR Clamd daemon not ready for stream scanning.\n");
	      fprintf (stdout, "\n");
              freeBuff(in_buff);
	      continue;
	    }
	  if (debug > 1)
	    logit(log_file, "DEBUG Received port %d from clamd.\n", port);

	  /* connect to clamd given port */
	  if ((wsockd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	    {
	      logit(log_file, "ERROR Can't create the Clamd socket.\n");
	      fprintf (stdout, "\n");
              freeBuff(in_buff);
	      continue;
	    }

	  server.sin_family = AF_INET;
	  server.sin_port = htons (port);

	  peer_size = sizeof (peer);

	  if (getpeername (sockd, (struct sockaddr *) &peer, (socklen_t *) &peer_size) < 0)
	    {
	      logit(log_file, "ERROR Can't get socket peer name.\n");
	      fprintf (stdout, "\n");
	      continue;
	    }
	  switch (peer.sin_family)
	    {
	    case AF_UNIX:
	      server.sin_addr.s_addr = inet_addr ("127.0.0.1");
	      break;
	    case AF_INET:
	      server.sin_addr.s_addr = peer.sin_addr.s_addr;
	      break;
	    default:
	      logit(log_file, "ERROR Unexpected socket type: %d.\n",
		   peer.sin_family);
              freeBuff(in_buff);
	      continue;
	    }

	  if (debug > 1)
	    logit(log_file, "DEBUG Trying to connect to clamd [port: %d].\n", port);

	  if (connect
	      (wsockd, (struct sockaddr *) &server,
	       sizeof (struct sockaddr_in)) < 0)
	    {
	      close (wsockd);
	      logit(log_file, "ERROR Can't connect to clamd [port: %d].\n", port);
	      fprintf (stdout, "\n");
              freeBuff(in_buff);
	      continue;
	    }

	  /***** now get the body content *****/
	  /* download body */
	  curl_easy_setopt (eh, CURLOPT_NOBODY, FALSE);
	  curl_easy_setopt (eh, CURLOPT_HTTPGET, TRUE);
	  if (curl_easy_perform (eh) != 0)
	    {
	      logit(log_file, "ERROR fail downloading url %s\n", in_buff->url);
	      logit(log_file, "ERROR CURLOPT_ERRORBUFFER: %s\n", err);
	      fprintf (stdout, "\n");
	      close (wsockd);
	      close (sockd);
	      if (debug > 1)
	        logit(log_file, "DEBUG Connection to clamd on port: %d closed.\n", port);
		if (statit)
			timeit(g_start, "Total");
              freeBuff(in_buff);
	      continue;
	    }
	  close (wsockd);
	  if (debug > 1)
	    logit(log_file, "DEBUG Connection to clamd on port: %d closed.\n", port);
	  if (debug > 1)
	    logit(log_file, "DEBUG Reading clamd scan result.\n");
	  memset(cbuff, 0, sizeof(cbuff));
	  memset(urlredir, 0, sizeof(urlredir));
	  virusfound = 0;
	  while ((nbread = read(sockd, cbuff, NORMAL_BUFF)) > 0)
	    {
	      if (debug > 1)
		logit(log_file, "DEBUG received from Clamd: %s", cbuff);
	      if (strstr (cbuff, "FOUND\n"))
		{
		  char *badurl = (char *)malloc(sizeof(char)*NORMAL_BUFF);
		  xstrncpy(badurl, in_buff->url, NORMAL_BUFF);
		  snprintf(urlredir, MAX_BUFF, "%s?url=%s&source=%s&user=%s&virus=%s",
			redirect_url,badurl,in_buff->src_address,in_buff->ident,cbuff);
		  xfree(badurl);
		  replace_chr(urlredir, " ", "+");
		  /* chop urlredir */
		  chomp(urlredir);
		  logit(log_file, "LOG Redirecting URL to: %s\n", urlredir);
		  virusfound = 1;
		}
	        memset(cbuff, 0, sizeof(cbuff));
	    }
	  close (sockd);
	  if (debug > 1)
	    logit(log_file, "DEBUG End reading clamd scan result.\n");
	  if (statit)
		timeit(cs_start, "Virus Scanning");

	  if (virusfound == 1) {
	     if (trust_cache == 1) {
	        if (debug != 0)
	           logit(log_file, "DEBUG Virus found, trying to purge Squid Cache.\n");
	        if (purge_cache(in_buff->url) == 0) {
	           trust_cache = 0;
	        } else {
	           logit(log_file, "Squid Cache purged of url %s.\n", in_buff->url);
	        }
	     }
	     fprintf (stdout, "%s %s %s %s\n", urlredir,
		       in_buff->src_address, in_buff->ident, in_buff->method);
	     memset(urlredir, 0, sizeof(urlredir));
	     if (debug != 0)
	        logit(log_file, "DEBUG Virus found, redirection sent to Squid.\n");
	  } else {
	      if (debug != 0)
		logit(log_file, "DEBUG No virus detected.\n");
	      fprintf (stdout, "\n");
	  }
	}
        freeBuff(in_buff);
	if (statit)
		timeit(g_start, "Total");

    }

  /* cleanup */
  xfree(command);
  xfree(escaped);
  xfree(sbuff);
  free_global();

  return 0;
}

void sighup ()
{
  signal (SIGHUP, sighup);	/* reset signal */
  logit(log_file, "LOG Sighup received, cleaning all before quit...\n");
  /* simply terminate the program, squid should reload it */
  /* cleanup */
  curl_slist_free_all(slist);
  curl_global_cleanup ();
  free_global();
  exit (0);
}

void sigterm ()
{
  signal (SIGTERM, sigterm);	/* reset signal */
  logit(log_file, "LOG Sigterm received, cleaning all before quit...\n");
  /* cleanup */
  curl_slist_free_all(slist);
  curl_global_cleanup ();
  free_global();
  exit (0);
}

void free_global ()
{
  xfree(log_file);
  xfree(proxy);
  xfree(clamd_port);
  xfree(clamd_ip);
  xfree(clamd_local);
  xfree(redirect_url);
  xfree(squidguard);
  xfree(alternate_config);
  xfree(useragent);
  regfree(regexv);
  curl_slist_free_all(slist);
  curl_global_cleanup ();
  if (sgfpw) fclose(sgfpw);
  if (sgfpr) fclose(sgfpr);
}

int dconnect ()
{
  struct sockaddr_un userver;
  int asockd;


  memset ((char *) &userver, 0, sizeof (userver));

  if (clamd_local != NULL)
    {
      userver.sun_family = AF_UNIX;
      xstrncpy (userver.sun_path, clamd_local, sizeof(userver.sun_path));
      if ((asockd = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
	{
	  logit(log_file, "ERROR Can't bind local socket on %s.\n", clamd_local);
	  return -1;
	}
      if (connect (asockd, (struct sockaddr *) &userver, sizeof (struct sockaddr_un)) < 0) {
	  close (asockd);
	  logit(log_file, "ERROR Can't connect to clamd on local socket %s.\n", clamd_local);
	  return -1;
	}

    }
  else
    {
	asockd = connectINET(clamd_ip, atoi(clamd_port));
	if ( asockd == -1 ) {
		logit(log_file,"ERROR: unable to connect to Clamd: %s\n", strerror(errno) );
		return 0;
	}
    }
  return asockd;
}

void replace_chr (char string[], char *from, char *to)
{
  char *start, *p1, *p2;
  for (start = string; *start != '\0'; start++)
    {
      p1 = from;
      p2 = start;
      while (*p1 != '\0')
	{
	  if (*p1 != *p2)
	    break;
	  p1++;
	  p2++;
	}
      if (*p1 == '\0')
	{
	  for (p1 = to; *p1 != '\0'; p1++)
	    *start++ = *p1;
	}
    }
}

void timeit (struct timeval start, char *level)
{
	struct timeval stop;

	/* Stop timing and compute elapsed time */
	gettimeofday(&stop,NULL);
	/* show stat */
	logit(log_file, "STAT %s process time %7.3f second(s)\n", level, (stop.tv_sec + (stop.tv_usec * 0.000001)) - (start.tv_sec + (start.tv_usec * 0.000001)));
}

/* 
   function to compute new string length,
   allocate memory and finally copy string 
   - Returned value must be freed -
 */
static char * escape_quote (char *s)
{
	char *newstring;
	char *idx1, *idx2;
	int nbquotes = 0;
	int len = 0;

	if (!s)
		return NULL;

	len = xstrnlen (s, MAX_BUFF);

	/* First, count quotes in string */
	idx1 = s;
	while (*idx1) {
		if ((*(idx1) == '"') || (*(idx1) == '\\') || (*(idx1) == '$'))
			nbquotes++;
		idx1++;
	}
	/* If there is no quote in the string, return it */
	if (!nbquotes)
		return (strdup (s));

	/* Or else, allocate memory for the new string */
	newstring = (char *) malloc (sizeof (char) * (len + nbquotes + 1));
	if (newstring == NULL) {
		logit(log_file, "PANIC Unable to allocate memory in escape_quote()\n");
		return NULL;
	}
	/* And prepend a backslash before any quote found in string */
	idx1 = s;
	idx2 = newstring;
	while (*idx1) {
		if ((*(idx1) == '"') || (*(idx1) == '\\') || (*(idx1) == '$')) {
			*(idx2++) = '\\';
		}
		*(idx2++) = *idx1++;
	}
	/* Add null char */
	*idx2 = '\0';

	return newstring;
}

void freeBuff (struct IN_BUFF * in_buff)
{
      xfree(in_buff->url);
      xfree(in_buff->src_address);
      xfree(in_buff->ident);
      xfree(in_buff->method);
      xfree(in_buff);
}

void sigusr1 ()
{
  signal (SIGUSR1, sigusr1);	/* reset signal */
  logit(log_file, "LOG Sigusr1 received, re-reading configuration...\n");
  load_patterns(1);
  regfree(regexv);
  regexv = regcomp_pattern();
  if (bridge_mode)
    logit(log_file, "PANIC Invalid condition - continuing in BRIDGE mode\n");

}

char *replace_str(char *str, char *orig, char *rep)
{
  static char buffer[MAX_BUFF];
  char *p;

  if(!(p = strstr(str, orig)))  // Is 'orig' even in 'str'?
    return str;

  strncpy(buffer, str, p-str); // Copy characters from 'str' start to 'orig' st$
  buffer[p-str] = '\0';

  sprintf(buffer+(p-str), "%s%s", rep, p+strlen(orig));

  return buffer;
}


int purge_cache(char *url)
{
	static const char* purge_cmd = "PURGE %s HTTP/1.0\r\nAccept: */*\r\n\r\n";
	char* buffer = NULL;
	unsigned long buff_size = strlen(url) + strlen(purge_cmd) + 4;

	buffer = (char *)malloc(buff_size);
	snprintf( buffer, buff_size, purge_cmd, url );
	int sockfd = connectINET(squid_ip, atoi(squid_port));
	if ( sockfd == -1 ) {
		logit(log_file,"ERROR: unable to connect to Squid: %s\n", strerror(errno) );
		xfree(buffer);
		return 0;
	}

	int size = strlen(buffer);
	if ( write( sockfd, buffer, size ) != size ) {
		logit(log_file,"ERROR: unable to talk to squid: %s\n", strerror(errno) );
		close(sockfd);
		xfree(buffer);
		return 0;
	}
	memset( buffer+8, 0, 4 );
	if ( read( sockfd, buffer, buff_size ) < 1 ) {
		logit(log_file,"ERROR: unable to read squid answer: %s\n", strerror(errno) );
		close(sockfd);
		xfree(buffer);
		return 0;
	}
	close(sockfd);
	int result = strtol(buffer+8,0,10);
	if (debug > 0)
	   logit(log_file,"DEBUG Purge squid answer: %d\n", result);
	if (result == 403) {
	      logit(log_file,"ERROR Purge method not allowed. Please edit your squid ACL.\n");
	      logit(log_file,"ERROR Fallback to untrusted cache.\n");
	      trust_cache = 0;
	}
	xfree(buffer);

	return 1;
}

int connectINET(char *serverHost, int serverPort)
{
	struct sockaddr_in server;
	struct hostent *he;
	int asockd;

	memset ((char *) &server, 0, sizeof (server));

	server.sin_addr.s_addr = inet_addr(serverHost);
	if ((asockd = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	{
	  logit(log_file, "ERROR Can't create a socket.\n");
	  return -1;
	}

	server.sin_family = AF_INET;
	server.sin_port = htons(serverPort);

	if ((he = gethostbyname(serverHost)) == 0)
	{
	  close (asockd);
	  logit(log_file, "ERROR Can't lookup hostname of %s\n", serverHost);
	  return -1;
	}
	server.sin_addr = *(struct in_addr *) he->h_addr_list[0];

	if (connect (asockd, (struct sockaddr *) &server, sizeof (struct sockaddr_in)) < 0) {
	  close (asockd);
	  logit(log_file, "ERROR Can't connect on server:port %s:%d.\n", serverHost,serverPort);
	  return -1;
	}
	return asockd;
}

